"""Various helper functions"""
import codecs
import datetime
import logging
import os
import os.path

import pandas as pd
import rich_click as click
from rich import print as rprint
from rich.logging import RichHandler
from rich.prompt import Prompt
from tabulate import tabulate

FORMAT = "%(message)s"
rhandler = RichHandler(
    level=logging.NOTSET,
    show_level=True,
    show_path=False,
    show_time=True,
    omit_repeated_times=False,
)

logging.basicConfig(level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[rhandler])
log = logging.getLogger("rich")


def combine_results(results_list):
    """
    Combine a list of dataframes containing results into a single dataframe
    """
    if not results_list:
        return pd.DataFrame()

    results = pd.concat(results_list, ignore_index=True)
    results = results.sort_values(
        by=["level_1", "level_2", "level_3", "level_4", "level_5"]
    ).reset_index(drop=True)

    return results


def write_results_to_file(results, output):
    """
    Write a dataframe containing results to a file in CSV or TSV format
    """
    if not output:
        return

    if not os.path.dirname(output):
        output = os.path.join(os.getcwd(), output)

    try:
        if os.path.exists(output):
            while True:
                user_input = Prompt.ask(
                    f"[yellow bold]File[/] [cyan italic]{output}[/] [yellow bold]already exists.\n"
                    f"Do you want to overwrite it? (y/n) [dim] Press [italic]ENTER[/] to exit",
                )
                if user_input.lower() == "y":
                    rprint(f"[yellow bold]Overwriting[/] [italic]{output}")
                    break
                elif user_input.lower() == "n":
                    now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                    new_output = f"{now}_{os.path.basename(output)}"
                    output_dir = os.path.dirname(output)
                    if not output_dir:
                        output_dir = os.getcwd()
                    new_output = os.path.join(output_dir, new_output)
                    new_output = os.path.splitext(new_output)[0] + ".tsv"
                    results.to_csv(new_output, index=False, sep="\t")
                    rprint(f"[yellow bold]Saving as[/] [italic]{new_output}")
                    return
                elif user_input == "":
                    raise KeyboardInterrupt
                else:
                    rprint(
                        "[red bold]Invalid input.[/] [yellow]Please enter [italic]'y'[/] or [italic]'n'[/]."
                    )

        os.makedirs(os.path.dirname(output), exist_ok=True)

        ext = os.path.splitext(output)[1]
        if ext == ".csv":
            results.to_csv(output, index=False)
        elif ext == ".tsv" or ext == ".txt":
            results.to_csv(output, index=False, sep="\t")
        else:
            raise SystemExit("Output file must have 'txt', 'tsv', or 'csv' extension")

    except OSError as e:
        if e.errno == 30:
            log.error(
                "File cannot be created at this path. Read-only file system. Please check your file system permissions."
            )
        else:
            raise e


def print_results(results):
    """
    Print a dataframe containing results to the console in a formatted table.

    Args:
    ____
        results (pandas.DataFrame): A DataFrame containing the results to be printed.

    Returns:
    _______
        None.
    """
    headers = ["Sample", "Level 1", "Level 2", "Level 3", "Level 4", "Level 5"]
    if results.empty:
        rprint("[red bold]ATTENTION[/]: [italic]No lineages were called!")
    else:
        table = tabulate(results, headers=headers, tablefmt="fancy_grid")
        print(f"\n{table}")


class InputOutputValidator:
    """Class to validate input and output"""

    def __init__(self, ctx, vcf_files, output):
        """
        Initialize InputOutputValidator.

        Parameters:
        -----------
        ctx : click.Context
            The click context object.
        vcf_files : tuple
            Tuple containing VCF files.
        output : str
            File path to write output to.
        """
        self.ctx = ctx
        self.vcf_files = vcf_files
        self.output = output

    def validate_input(self):
        """
        Validate input.

        Raises:
        -------
        click.exceptions.ClickException:
            If no VCF files are provided and help flag is not set.
        """
        if not self.vcf_files and not self.ctx.params.get("help"):
            rprint(
                "[red bold]ATTENTION[/]: [italic]Please provide one or more VCF files to process"
            )
            click.echo(self.ctx.get_help())
            self.ctx.exit()

    def validate_output(self):
        """
        Validate output.

        Raises:
        -------
        click.exceptions.ClickException:
            If output file doesn't have valid extension or help flag is set.
        """
        if self.output and not self.output.endswith((".txt", ".tsv", ".csv")):
            rprint(
                "[red bold]ATTENTION[/]: [italic]Output file must have 'txt', 'tsv', or 'csv' extension",
            )
            click.echo(self.ctx.get_help())
            self.ctx.exit()

        elif self.output:
            rprint(f"[yellow bold]Writing results to[/] [cyan italic]{self.output}[/]")


def read(rel_path):
    """
    Read file at the specified path.

    Parameters:
    -----------
    rel_path : str
        Path to the file relative to this one.

    Returns:
    --------
    str:
        Contents of the file.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    """
    Extract version number from the specified file.

    Parameters:
    -----------
    rel_path : str
        Path to the file relative to this one.

    Returns:
    --------
    str:
        The version number.
    """
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")



"""Process levels.tsv file"""
from importlib.resources import files

import pandas as pd


def get_levels_data():
    """
    Read the data from a TSV file containing information about barcoding levels, and extract the relevant data
    for each level. The data for each level is then stored in a tuple of NumPy arrays, where each array
    contains the position, the reference allele, the alternative allele, and the lineage for each row in the data.
    Additionally, the function returns an array of positional data that is common to all levels.


    Returns:
    -------
    levels_data : Tuple[Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray]
        A tuple of NumPy arrays, where each sub-tuple contains the data for a different level. Each sub-tuple
        contains four arrays representing the position, the reference allele, the alternative allele,
        and the lineage for each row in the data.
    pos : np.ndarray
        An array containing the position for all SNPs.
    """
    levels_file = files("tblg.data").joinpath("levels.tsv")
    temp_df = pd.read_csv(levels_file, sep="\t")
    unique_levels = temp_df["level"].unique()
    levels_dict = {}

    # For each unique level value, extract the relevant data and store it in the dictionary
    for elem in unique_levels:
        levels_dict[elem] = temp_df[temp_df["level"] == elem][
            ["POS", "REF", "ALT", "lineage"]
        ]

    # Create a tuple of NumPy arrays, where each array contains the data for a different level
    # Each array contains the positional data, the reference allele, the alternative allele,
    # and the lineage (i.e., the level value) for each row in the data
    # The tuple contains data for levels 1 through 5
    levels_data = tuple(
        levels_dict.get(i, pd.DataFrame())[["POS", "REF", "ALT", "lineage"]]
        .to_numpy()
        .T
        for i in range(1, 6)
    )

    pos = temp_df["POS"].values

    return levels_data, pos



"""
TbLG (Tuberculosis Lineage Genotyping)

A command-line tool to genotype Mycobacterium tuberculosis lineage from a VCF file

Author: Dmitry Bespiatykh
"""
import sys

import rich_click as click
from rich import print as rprint

from .barcoding import process_vcf_files
from .utils import (
    InputOutputValidator,
    combine_results,
    get_version,
    print_results,
    write_results_to_file,
)

version = get_version("__init__.py")

click.rich_click.MAX_WIDTH = 75
click.rich_click.USE_RICH_MARKUP = True
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.HEADER_TEXT = f"[blue bold]TbLG[/][dim] (Tuberculosis Lineage Genotyping)[cyan] |[/] [cyan bold]v{version}[/]"
click.rich_click.ERRORS_SUGGESTION = f"[blue bold]TbLG[/] [cyan bold]v{version}[/]\nRun '[cyan]tblg --help or -h[/]' to show help message."
click.rich_click.STYLE_ERRORS_SUGGESTION = "cyan italic"
click.rich_click.STYLE_OPTION = "cyan"

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "vcf_files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True),
    metavar="[VCF FILES]",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    help="Write results to file [dim]['.txt', '.tsv', or '.csv']",
)
@click.version_option(version, "-v", "--version", is_flag=True)
@click.pass_context
def main(ctx, vcf_files, output):
    """
    Process one or more VCF files and genotype lineages.
    """
    validator = InputOutputValidator(ctx, vcf_files, output)
    validator.validate_input()
    validator.validate_output()

    results_list = process_vcf_files(vcf_files)

    if not results_list:
        rprint("[red bold]ATTENTION[/]: [italic]No valid VCF files were found!")
        return

    results = combine_results(results_list)

    if output:
        write_results_to_file(results, output)
    else:
        print_results(results)


if __name__ == "__main__":
    sys.exit(main())



"""Barcoding related functions"""
import numpy as np
from tqdm.auto import tqdm

from .levels import get_levels_data
from .vcf import validate_vcf, vcf_to_dataframe


def lineage2_decision(call_list):
    """
    Determines the lineage 2 variant calls and returns a modified call list.

    Args:
    call_list (list): A list of variant calls.

    Returns:
    altList (list): A modified list of variant calls, where any call containing
                    both "L2.2 (modern)" and "L2.2 (ancient)" is modified to
                    contain only "L2.2 (modern)" and any call not containing
                    "L2.2 (modern)" is left unchanged.
    """
    lin2 = ["L2.2 (modern)", "L2.2 (ancient)"]
    alt_list = []
    for item in call_list:
        if all(i in item for i in lin2):
            item = list(set(item) - set(lin2))
            item.append(lin2[0])
        alt_list.append(item)
    return alt_list


def lineage4_decision(call_list, lin):
    """
    Determines the alternate alleles for lineage 4 based on the input call list and lineage 4 variants.
    This function takes a list of calls and a lineage as input and returns a new list
    that either removes the lineage if it exists in the call or adds the lineage to the
    call twice if it does not exist in the call.

    Args:
        call_list (List[List[str]]): A list of lists containing genotype calls for each sample at a given position.
        lin (List[str]): A list of lineage 4 variants.

    Returns:
        List[List[str]]: A list of lists containing the alternate alleles for each sample at a given position.
    """
    alt_list = []
    for item in call_list:
        if any(i in item for i in lin):
            item = [x for x in item if x not in lin]
        else:
            item.extend([lin[0] for i in range(2)])
        alt_list.append(item)
    return alt_list


def count_variants(call_list, prefix=None):
    """
    Count the number of occurrences of each variant in the input list and modify it as follows:
    - If a variant occurs more than once, keep it as-is.
    - If a variant occurs only once, concatenate it with a warning message in square brackets.
    - If a variant starts with the prefix string (if provided), remove it from the list.

    Args:
        call_list (List[str]): A list of variants.
        prefix (Optional[str]): A string prefix to remove from the list.

    Returns:
        List[str]: A modified version of the input list, with variants possibly concatenated with a warning message or removed.
    """
    d = {}
    for item in call_list:
        if item:
            caseless = item.casefold()
            if caseless in d:
                d[caseless][1] += 1
            else:
                d[caseless] = [item, 1]

    call_list = []
    for item, count in d.values():
        if not item.startswith(prefix) if prefix else True:
            item = f"{item}" if count > 1 else f"{item}*"
        call_list.append(item)

    return call_list


def barcoding(uploaded_vcf, use_tqdm=False):
    """
    Perform barcoding of a VCF file uploaded by the user.

    Parameters:
    -----------
    uploaded_vcf : str
        The file path or name of the uploaded VCF file.
    use_tqdm : bool, optional
        Whether or not to use tqdm to display progress bar (default is False).

    Returns:
    --------
    pandas.DataFrame
        A dataframe containing barcoding analysis results for each sample.
    """
    # Convert VCF to DataFrame
    df = vcf_to_dataframe(uploaded_vcf, use_tqdm=use_tqdm)

    # Get levels dictionary and create a list of level names
    levels = get_levels_data()[0]
    level_names = [f"level_{i+1}" for i in range(len(levels))]

    # Define a function to compute the level of each sample
    def compute_level(level):
        """
        The compute_level function takes a level tuple and computes the level for
        each sample in a DataFrame based on the values of "POS", "REF", and "ALT"
        columns of the DataFrame. It returns an array of levels for each sample.

        Parameters:
        -----------
        level: A tuple containing four elements
            - the position (POS), reference allele (REF), alternate allele (ALT)
            and the level number. This tuple represents a single level.
        Returns:
        --------
        levels: A numpy array containing the level number for each sample in the DataFrame.
        Note: This function assumes that the input DataFrame df has columns named "POS", "REF", and "ALT".
        """
        exp = (
            df["POS"].values[:, None],
            df["REF"].values[:, None],
            df["ALT"].values[:, None],
        )
        mask = np.logical_and.reduce(
            [
                np.equal(exp[0], level[0]),
                np.equal(exp[1], level[1]),
                np.equal(exp[2], level[2]),
            ]
        )
        return np.dot(mask, level[3])

    # Compute the level of each sample for each level
    for i, level in (
        enumerate(tqdm(levels, desc="Processing levels", colour="blue"))
        if use_tqdm
        else enumerate(levels)
    ):
        df[level_names[i]] = compute_level(level)

    # Drop the columns REF, ALT, and POS and replace empty strings with NaN
    df.drop(["REF", "ALT", "POS"], axis=1, inplace=True)
    df.replace("", np.nan, inplace=True)

    # Group the data by sample and concatenate the level columns into comma-separated strings
    df = df.groupby(["Sample"]).agg(lambda x: ",".join(x.dropna())).reset_index()

    # Split the first two level columns into lists and apply lineage decision and count variants functions
    df[level_names[:2]] = df[level_names[:2]].applymap(lambda x: x.split(","))
    df[level_names[0]] = lineage4_decision(df[level_names[0]], ["L4"])
    df[level_names[1]] = lineage4_decision(df[level_names[1]], ["L4.9"])
    df[level_names[0]] = df[level_names[0]].apply(count_variants, prefix="L8")
    df[level_names[1]] = df[level_names[1]].apply(
        count_variants, prefix=("L2.2 (modern)", "L2.2 (ancient)")
    )
    df[level_names[1]] = lineage2_decision(df[level_names[1]])

    # Convert the first two level columns back to comma-separated strings
    df[level_names[:2]] = df[level_names[:2]].applymap(lambda x: ", ".join(map(str, x)))

    # Sort the dataframe by level and reset the index
    df.sort_values(level_names, inplace=True)
    df.reset_index(drop=True, inplace=True)

    # Return the final dataframe
    return df


def process_vcf_files(vcf_files):
    """
    Process one or more VCF files and return a list of dataframes containing results
    """
    results_list = []

    # Check if only one VCF file is supplied
    if len(vcf_files) == 1:
        vcf_file = vcf_files[0]
        if validate_vcf(vcf_file):
            out = barcoding(vcf_file, use_tqdm=True)
            results_list.append(out)

    # If more than one VCF file is supplied, process each file and display a progress bar
    else:
        with tqdm(total=len(vcf_files), desc="Processing files", colour="blue") as pbar:
            for vcf_file in vcf_files:
                if validate_vcf(vcf_file):
                    out = barcoding(vcf_file, use_tqdm=False)
                    if (
                        out.empty
                        or all(
                            out.loc[:, out.columns != "Sample"]
                            .replace("", np.nan)
                            .isna()
                            .all()
                        )
                        is True
                    ):
                        tqdm.write(
                            f"{vcf_file} does not have any genotyping SNPs. Skipping..."
                        )
                    else:
                        results_list.append(out)
                pbar.update(1)

    return results_list



import codecs
import os.path

from setuptools import find_namespace_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="tblg",
    version=get_version("src/tblg/__init__.py"),
    description="A command-line tool to genotype Mycobacterium tuberculosis lineage from a VCF file",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dbespiatykh/tblg",
    author="Dmitry Bespiatykh",
    author_email="d.bespiatykh@gmail.com",
    license="MIT",
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={"tblg.data": ["*.tsv"]},
    entry_points={"console_scripts": ["tblg = tblg.__main__:main"]},
    install_requires=[
        "click>=8.0.4",
        "numpy>=1.23.5",
        "pandas>=1.5.0",
        "tabulate>=0.9.0",
        "tqdm>=4.64.1",
        "rich>=13.3.1",
        "rich-click>=1.6.1",
        "datetime>=5.1",
    ],
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



"""Process VCF files"""
from gzip import open as gzopen

import numpy as np
import pandas as pd
from tqdm import tqdm

from .levels import get_levels_data


def validate_vcf(vcf_file):
    """
    Check if a VCF file has typical VCF structure.

    Returns True if the file has typical VCF structure, and False otherwise.
    """
    if vcf_file.endswith(".vcf.gz"):
        with gzopen(vcf_file, "rb") as f:
            for line in f:
                line = line.decode("utf-8")
                if line.startswith("#"):
                    continue
                else:
                    fields = line.strip().split("\t")
                    if len(fields) < 8:
                        tqdm.write(
                            f"{vcf_file} does not have typical VCF structure. Skipping..."
                        )
                        return False
                    else:
                        return True
    elif vcf_file.endswith(".vcf"):
        with open(vcf_file) as f:
            for line in f:
                if line.startswith("#"):
                    continue
                else:
                    fields = line.strip().split("\t")
                    if len(fields) < 8:
                        tqdm.write(
                            f"{vcf_file} does not have typical VCF structure. Skipping..."
                        )
                        return False
                    else:
                        return True
    else:
        tqdm.write(f"{vcf_file} does not end with .vcf or .vcf.gz. Skipping...")


def vcf_to_dataframe(file, use_tqdm=False):
    """
    Reads a VCF file and returns a Pandas DataFrame containing the sample names, position,
    and alternative allele for each sample in the VCF file.

    Parameters:
    ----------
    file : str
        The name of the VCF file to read. The file can be compressed with gzip.
    use_tqdm : bool, optional
        A flag that indicates whether or not to display a progress bar during the reading process.
        Default is False.

    Returns:
    -------
    df : pd.DataFrame
        A Pandas DataFrame containing the sample names, positional data, and allele data for each sample
        in the VCF file. The DataFrame is filtered to only include rows with positions that are present in
        the levels data extracted from the levels.tsv file.
    """
    pos_all = get_levels_data()[1]

    opener = gzopen if file.endswith(".gz") else open

    with opener(file, "rt") as f:
        # Remove all header lines from the input file
        lines = [line.strip() for line in f if not line.startswith("##")]

    # Extract the sample names from the first header line
    header = lines.pop(0).split("\t")[9:]

    # Extract data from each line of the input file and convert it to a list
    data = []

    if use_tqdm:
        pbar = tqdm(total=len(lines), desc="Reading VCF files", colour="blue")
    for i, line in enumerate(lines):
        fields = line.split("\t")
        pos, ref, alt = fields[1], fields[3], fields[4]
        alleles = [ref] + alt.split(",")
        genotypes = [genotype.split(":") for genotype in fields[9:]]

        for i, gt in enumerate(genotypes):
            if gt[0] == "." or gt[0] == "./." or gt[0] == ".|.":
                allele = np.nan
            elif "/" in gt[0]:
                alleles_list = [
                    alleles[int(x)] if x != "." else np.nan for x in gt[0].split("/")
                ]
                allele = "/".join([x for x in alleles_list if str(x) != "nan"])
            elif "|" in gt[0]:
                alleles_list = [
                    alleles[int(x)] if x != "." else np.nan for x in gt[0].split("|")
                ]
                allele = "|".join([x for x in alleles_list if str(x) != "nan"])
            else:
                allele = alleles[int(gt[0])]
            data.append([header[i], pos, ref, allele])
        if use_tqdm:
            pbar.update(1)
    if use_tqdm:
        pbar.close()

    # Convert the list of data to a Pandas DataFrame
    df = pd.DataFrame(data, columns=["Sample", "POS", "REF", "ALT"])

    # Set the data types for each column in the DataFrame
    df = df.astype(
        {"Sample": "object", "REF": "object", "ALT": "object", "POS": "int64"}
    )
    # Strip "/" and "|" from ALT column
    df["ALT"] = df["ALT"].str.split(r"[/|]").str[-1]

    # Filter the DataFrame to only include rows with positions in the pos_all list
    df = df[df["POS"].isin(pos_all)].reset_index(drop=True)
    # Return the filtered DataFrame
    return df
